X = 1
Y = 2
SIZE = 3
ICONS = "　ＸＯ"
ICON_SELECTED = "＊"
VERT_LINE = "｜"
HORI_LINE = "ー"
CROSS = "＋"

class CountResult:
    def __init__(self, player, v):
        self.player = player
        self.v = v

    def then(self, c, i, p=True):
        if self.v or not p:
            return self
        return c.hit(self.player, i)

class Counts:
    def __init__(self, s):
        self.v = ([0]*s, [0]*s)

    def hit(self, player, i):
        self.v[player-1][i] += 1
        return CountResult(player, self.v[player-1][i] == SIZE)

    def dump(self):
        return self.v

    @classmethod
    def load(cls, d):
        me = cls(0)
        me.v = d
        return me

class TicTacToe:
    def __init__(self):
        self.board = [[0]*SIZE for _ in range(SIZE)]
        self.pipe_counts = Counts(SIZE)
        self.dash_counts = Counts(SIZE)
        self.reverse_solidus_counts = Counts(1)
        self.solidus_counts = Counts(1)

    def make_move(self, player, x, y):
        if not (0 <= x < SIZE and 0 <= y < SIZE):
            raise ValueError("invalid coordinates")
        if self.board[y][x]:
            raise ValueError("position not empty")

        self.board[y][x] = player
        # what a moment
        return (
            self.pipe_counts.hit(player, x)
            .then(self.dash_counts, y)
            .then(self.reverse_solidus_counts, 0, x == y)
            .then(self.solidus_counts, 0, SIZE-1-x == y)
        ).v

    def dump(self):
        return {
            "board": self.board,
            "pipe_counts": self.pipe_counts.dump(),
            "dash_counts": self.dash_counts.dump(),
            "reverse_solidus_counts": self.reverse_solidus_counts.dump(),
            "solidus_counts": self.solidus_counts.dump(),
        }

    @classmethod
    def load(cls, d):
        me = cls()
        me.board = d["board"]
        me.pipe_counts = Counts.load(d["pipe_counts"])
        me.dash_counts = Counts.load(d["dash_counts"])
        me.reverse_solidus_counts = Counts.load(d["reverse_solidus_counts"])
        me.solidus_counts = Counts.load(d["solidus_counts"])
        return me

class UltimateTicTacToe:
    def __init__(self):
        # this board has the winning state of the smaller boards. when it's won the game is over
        self.overall = TicTacToe()
        self.innards = [[TicTacToe() for _ in range(SIZE)] for _ in range(SIZE)]
        self.to_play = X
        self.next_space = None

    def is_virgin(self):
        for sy in self.innards:
            for sx in sy:
                for y in sx.board:
                    for x in y:
                        if x == Y:
                            return False
        return True

    def make_move(self, x, y, sx, sy):
        if not (0 <= x < SIZE and 0 <= y < SIZE):
            raise ValueError("invalid coordinates")
        if self.next_space and (x, y) != self.next_space:
            raise ValueError("sub-board not active")
        if self.overall.board[y][x]:
            raise ValueError("sub-board already completed")
        won = self.innards[y][x].make_move(self.to_play, sx, sy)
        try:
            return won and self.overall.make_move(self.to_play, x, y)
        finally:
            self.next_space = None if self.overall.board[sy][sx] else (sx, sy)
            self.to_play = Y if self.to_play == X else X

    def render(self, select):
        outside = ""
        out = ""
        # ok, here we go
        for sy in range(SIZE):
            for y in range(SIZE):
                for sx in range(SIZE):
                    for x in range(SIZE):
                        i = ICONS[self.innards[sy][sx].board[y][x]]
                        if (sx, sy) == select:
                            out += ICON_SELECTED
                            outside += i if i != ICONS[0] else "・"
                        else:
                            out += i
                    if sx < SIZE-1:
                        out += VERT_LINE
                out += "\n"
            if sy < SIZE-1:
                out += CROSS.join([HORI_LINE*SIZE]*SIZE) + "\n"
        return out + ("\n===\n" + outside[0:3] + "\n" + outside[3:6] + "\n" + outside[6:9] if outside else "")

    def dump(self):
        return {
            "overall": self.overall.dump(),
            "innards": [[x.dump() for x in y] for y in self.innards],
            "to_play": self.to_play,
            "next_space": self.next_space,
        }

    @classmethod
    def load(cls, d):
        me = cls()
        me.overall = TicTacToe.load(d["overall"])
        me.innards = [[TicTacToe.load(x) for x in y] for y in d["innards"]]
        me.to_play = d["to_play"]
        me.next_space = d["next_space"] and tuple(d["next_space"])
        return me
